"""Pipeline Step to extract networking traffic via nPrint: Net"""
import argparse
import itertools
import os
import pathlib
import re
import sys
import textwrap
import typing

import nprintml
from nprintml import pipeline

from .execute import nprint


class NetResult(typing.NamedTuple):
    """Pipeline Step results for Net"""
    nprint_path: pathlib.Path


class Net(pipeline.Step):
    """Extend given `ArgumentParser` with nPrint interface and invoke
    `nprint` command to initiate nprintML pipeline.

    Returns a `NetResult`.

    """
    def __init__(self, parser):
        self.group_parser = parser.add_argument_group(
            "extraction of features from network traffic via nPrint",

            "Full information can be found at https://nprint.github.io/nprint/"
        )

        self.group_parser.add_argument(
            '-4', '--ipv4',
            action='store_true',
            help="include ipv4 headers",
        )
        self.group_parser.add_argument(
            '-6', '--ipv6',
            action='store_true',
            help="include ipv6 headers",
        )
        self.group_parser.add_argument(
            '-A', '--absolute-timestamps', '--absolute_timestamps',
            action='store_true',
            help="include absolute timestamp field",
        )
        self.group_parser.add_argument(
            '-c', '--count',
            metavar='INTEGER',
            type=int,
            help="number of packets to parse (if not all)",
        )
        self.group_parser.add_argument(
            '-C', '--csv-file', '--csv_file',
            type=FileAccessType(os.R_OK),
            metavar='FILE',
            help="csv (hex packets) infile",
        )
        self.group_parser.add_argument(
            '-d', '--device',
            help="device to capture from if live capture",
        )
        self.group_parser.add_argument(
            '-e', '--eth',
            action='store_true',
            help="include eth headers",
        )
        self.group_parser.add_argument(
            '-f', '--filter',
            help="filter for libpcap",
        )
        self.group_parser.add_argument(
            '-i', '--icmp',
            action='store_true',
            help="include icmp headers",
        )
        self.group_parser.add_argument(
            '-I', '--ip-file', '--ip_file',
            metavar='FILE',
            type=FileAccessType(os.R_OK),
            help="file of IP addresses to filter with (1 per line), "
                 "can be combined with num_packets for num_packets per ip",
        )
        self.group_parser.add_argument(
            '-N', '--nprint-file', '--nPrint_file',
            metavar='FILE',
            type=FileAccessType(os.R_OK),
            help="nPrint infile",
        )
        self.group_parser.add_argument(
            '-O', '--write-index', '--write_index',
            choices=range(5),
            metavar='INTEGER',
            type=int,
            help=textwrap.dedent("""\
                output file index (first column)
                select from:
                    0: source IP (default)
                    1: destination IP
                    2: source port
                    3: destination port
                    4: flow (5-tuple)"""),
        )
        self.group_parser.add_argument(
            '-p', '--payload',
            metavar='INTEGER',
            type=int,
            help="include n bytes of payload",
        )
        self.group_parser.add_argument(
            '-P', '--pcap-file', '--pcap_file',
            default=(),
            metavar='FILE',
            nargs='*',
            type=FileAccessType(os.R_OK),
            help="pcap infile",
        )
        self.group_parser.add_argument(
            '--pcap-dir', '--pcap_dir',
            default=(),
            metavar='DIR',
            nargs='*',
            type=DirectoryAccessType(ext='.pcap'),
            help="directory containing pcap infile(s) with file extension '.pcap'",
        )
        self.group_parser.add_argument(
            '-R', '--relative-timestamps', '--relative_timestamps',
            action='store_true',
            help="include relative timestamp field",
        )
        self.group_parser.add_argument(
            '-t', '--tcp',
            action='store_true',
            help="include tcp headers",
        )
        self.group_parser.add_argument(
            '-u', '--udp',
            action='store_true',
            help="include udp headers",
        )

    @staticmethod
    def get_output_directory(args):
        return args.outdir / 'nprint'

    @classmethod
    def make_output_directory(cls, args):
        outdir = cls.get_output_directory(args)
        outdir.mkdir()
        return outdir

    @staticmethod
    def make_output_path(outdir, pcap_file, dir_basis):
        if not pcap_file:
            return None

        npt_file = pathlib.Path(pcap_file).with_suffix('.npt')
        if dir_basis:
            npt_file = npt_file.relative_to(dir_basis)
        else:
            npt_file = npt_file.name

        npt_path = outdir / npt_file

        if npt_path.exists():
            raise FileExistsError(None, 'nPrint output path collision', str(npt_path))

        npt_path.parent.mkdir(parents=True, exist_ok=True)
        return npt_path

    def generate_argv(self, args, pcap_file=None, npt_file=None):
        """Construct arguments for `nprint` command."""
        # generate shared/global arguments
        if args.verbose:
            yield '--verbose'

        # support arbitrary pcap infile(s)
        if pcap_file:
            yield from ('--pcap_file', pcap_file)

        # add group (nPrint-specific) arguments
        for action in self.group_parser._group_actions:
            if action.dest in ('pcap_file', 'pcap_dir'):
                continue

            key = action.option_strings[-1]
            value = getattr(args, action.dest)

            if value is not action.default:
                yield key

                if not isinstance(value, bool):
                    yield str(value)

        # add output path
        outdir = self.get_output_directory(args)
        outpath = npt_file or outdir / 'netcap.npt'
        yield from ('--write_file', str(outpath))

    def __call__(self, args, results):
        try:
            warn_version_mismatch()
        except nprint.NoCommand:
            args.__parser__.error("nprint command could not be found on PATH "
                                  "(to install see nprint-install)")

        outdir = self.make_output_directory(args)

        if args.pcap_file or args.pcap_dir:
            # stream pair of pcap path & "basis" for reconstructing tree
            pcap_files = itertools.chain(
                zip(args.pcap_file, itertools.repeat(None)),
                itertools.chain.from_iterable(
                    zip(pcap_dir.rglob('*.pcap'), itertools.repeat(pcap_dir))
                    for pcap_dir in args.pcap_dir
                ),
            )
        else:
            pcap_files = ((None, None),)

        for (pcap_file, dir_basis) in pcap_files:
            npt_file = self.make_output_path(outdir, pcap_file, dir_basis)
            nprint(
                *self.generate_argv(args, pcap_file, npt_file),
            )

        return NetResult(outdir)


def warn_version_mismatch():
    """Warn if nPrint intended version doesn't match what's on PATH."""
    version_result = nprint('--version', stdout=nprint.PIPE)
    version_output = version_result.stdout.decode()
    version_match = re.match(r'nprint ([.\d]+)', version_output)
    if version_match:
        version_installed = version_match.group(1)
        if version_installed != nprintml.__nprint_version__:
            command_path = version_result.args[0]
            print(
                "[warn]",
                f"nprint expected version for nprintML ({nprintml.__nprint_version__}) "
                f"does not match version on PATH ({version_installed} at {command_path})",
                file=sys.stderr,
            )
    else:
        print(
            "[warn]",
            f"failed to parse version of nprint installed ({version_output})",
            file=sys.stderr,
        )


class FileAccessType:
    """Argument type to test a supplied filesystem path for specified
    access.

    Access level is indicated by bit mask.

    `argparse.FileType` may be preferred when the path should be opened
    in-process. `FileAccessType` allows for greater flexibility -- such
    as passing the path on to a subprocess -- while still validating
    access to the path upfront.

    """
    modes = {
        os.X_OK: 'execute',
        os.W_OK: 'write',
        os.R_OK: 'read',
        os.R_OK | os.X_OK: 'read-execute',
        os.R_OK | os.W_OK: 'read-write',
        os.R_OK | os.W_OK | os.X_OK: 'read-write-execute',
    }

    def __init__(self, access):
        self.access = access

        if access not in self.modes:
            raise ValueError("bad mask", access)

    @property
    def mode(self):
        return self.modes[self.access]

    def __call__(self, path):
        if os.access(path, self.access):
            return path

        raise argparse.ArgumentTypeError(f"can't open '{path}' ({self.mode})")


class DirectoryAccessType:
    """Argument type to test a supplied filesystem directory path."""

    def __init__(self, *, ext='', exists=None, empty=False, non_empty=False):
        if ext:
            non_empty = True

        if non_empty:
            exists = True

        if empty and non_empty:
            raise TypeError("directory cannot be both empty and non-empty")

        self.ext = ext
        self.exists = exists
        self.empty = empty
        self.non_empty = non_empty

    def __call__(self, value):
        path = pathlib.Path(value)

        if self.exists is not None:
            if self.exists:
                if not path.is_dir():
                    raise argparse.ArgumentTypeError(f"no such directory '{value}'")
            else:
                if path.exists():
                    raise argparse.ArgumentTypeError(f"path already exists '{value}'")

                if not os.access(path.parent, os.W_OK):
                    raise argparse.ArgumentTypeError(f"path not write-accessible '{value}'")

        if self.empty and any(path.glob('*')):
            raise argparse.ArgumentTypeError(f"directory is not empty '{value}'")

        if self.non_empty:
            count = 0
            for (count, child) in enumerate(path.rglob('*' + self.ext), 1):
                if not os.access(child, os.R_OK):
                    raise argparse.ArgumentTypeError(f"path(s) not read-accessible '{child}'")

            if count == 0:
                raise argparse.ArgumentTypeError("directory has no contents " +
                                                 (f"({self.ext}) " if self.ext else "") +
                                                 f"'{value}'")

        return path



"""Subprocessing interface to nPrint command"""
import shutil
import subprocess


def nprint(*args, stdin=None, stdout=None, stderr=None, check=True):
    """Execute the `nprint` command via `subprocess`.

    Argumentation reflects that of `subprocess.run`.

    The `PIPE` sentinel is made available via the callable object:

        nprint.PIPE

    If the `nprint` command cannot be found on the PATH, then
    `NoCommand` is raised. This exception is also made available on the
    callable object --

        nprint.NoCommand

    -- as is `CommandError`, raised from the `subprocess` package's
    `CalledProcessError`.

    Returns the resulting `subprocess.CompletedProcess`.

    """
    # ensure we know what we're running via which
    cmd = shutil.which('nprint')

    if cmd is None:
        raise NoCommand

    try:
        return subprocess.run(
            (cmd,) + args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            check=check,
        )
    except subprocess.CalledProcessError as exc:
        raise CommandError from exc


class nPrintError(Exception):
    pass


class CommandError(nPrintError):
    pass


class NoCommand(nPrintError):
    pass


nprint.CommandError = CommandError
nprint.NoCommand = NoCommand
nprint.PIPE = subprocess.PIPE



import pathlib

import numpy as np
import pandas as pd

from . import LabelAggregator


class IndexLabelAggregator(LabelAggregator):
    """A LabelAggregator to aggregates labels according to the *index*
    of nPrint output.

    This can be the src / dst ip, the src / dst port, or an entire flow
    label.

    """
    def __init__(self, npt_csv, label_csv):
        if isinstance(npt_csv, str):
            npt_csv = pathlib.Path(npt_csv)

        if isinstance(npt_csv, pathlib.Path) and npt_csv.is_dir():
            try:
                (npt_csv,) = npt_csv.iterdir()
            except ValueError:
                raise OSError(
                    f"{self.__class__.__name__} expects exactly one nPrint data file "
                    f"but specified directory contains none or more than one: '{npt_csv}'"
                )

        super().__init__(npt_csv, label_csv)

    def __call__(self, compress=False, sample_size=1):
        """Generates index driven features by loading nPrints and
        attaching labels, grouping by a given sample size if desired.

        """
        print('Loading nPrint:', self.filerepr(self.npt_csv))
        npt = self.load_npt(self.npt_csv)

        print('Loaded 1 nprint')
        print('  nPrint shape:', npt.shape)

        if compress:
            print('Compressing nPrint')
            npt = self.compress_npt(npt)
            print('  compressed nPrint shape:', npt.shape)

        print('Loading labels:', self.filerepr(self.label_csv))
        labels = self.load_label(self.label_csv)
        print('  number of labels:', labels.shape[0])

        if sample_size > 1:
            print('Grouping by sample size')
            npt = self.regroup_npt(npt, sample_size)
            print('  New shape of dataframe:', npt.shape)

        print('Attaching labels to nPrints')
        (npt, missing_labels, ogns, nns) = self.attach_label(npt, labels)
        print('  labels attached: missing labels for:')
        for missing_label in missing_labels:
            print('    {0}'.format(missing_label))
        print('    missing labels caused samples to be dropped:', (ogns - nns))

        return npt

    def regroup_npt(self, npt, sample_size):
        """Take a large nPrint dataframe, groups it by the index, then
        split each separated sample into n sample_sized samples.

        """
        # If the sample size is 1, its just per packet, which we already have
        if sample_size <= 1:
            return npt

        grouped = []
        indexes = []

        for (lid, group) in npt.groupby(npt.index):
            # Pad rows to get an even division for sample size
            pad_rows = sample_size - group.shape[0] % sample_size
            fill_row = [-1] * group.shape[1]
            for index in range(group.shape[0], group.shape[0] + pad_rows):
                group.loc[index] = fill_row

            # actually separate group into samples and flatten for 1d learning
            num_splits = group.shape[0] / sample_size
            for sample in np.split(group, num_splits):
                flattened_sample = sample.to_numpy().flatten()
                grouped.append(flattened_sample)
                indexes.append(lid)

        cols_flat = self.flatten_columns(npt.columns, sample_size)
        return pd.DataFrame(grouped, index=indexes, columns=cols_flat)



"""nPrint installation bootstrapping command"""
import argparse
import enum
import os
import pathlib
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
import urllib.request

import argparse_formatter

import nprintml


NPRINT_RELEASE_NAME = f'nprint-{nprintml.__nprint_version__}'

NPRINT_ARCHIVE_NAME = f'{NPRINT_RELEASE_NAME}.tar.gz'

NPRINT_ARCHIVE_URL = (
    'https://github.com/nprint/nprint/releases/download/'
    f'v{nprintml.__nprint_version__}/{NPRINT_ARCHIVE_NAME}'
)

USER_LOCAL_PATH = pathlib.Path.home() / '.local'

PYENV_PATH = os.getenv('PYENV_ROOT') or pathlib.Path.home() / '.pyenv'

BUILD_COMMANDS = (
    ('sh', 'configure', '--prefix={prefix}'),
    ('make',),
    ('make', 'install'),
)

BUILD_DEPENDENCIES = (
    'pcap',
)


class PathOption(str, enum.Enum):
    """system installation path options"""

    system_global = ('system-globally', None)
    current_user = ('for the current user only', USER_LOCAL_PATH)
    same_environ = ('into the same environment as nprintML', sys.prefix)
    user_specified = ('under another arbitrary path', None)

    def __new__(cls, description, target):
        # we really want to treat these as str (description)
        # (with attributes set elsewhere)
        obj = super().__new__(cls, description)
        obj._value_ = description
        return obj

    def __init__(self, description, target):
        self.description = description
        self.target = target

    @classmethod
    def get_path_defaults(cls):
        """Select most appropriate default installation path given
        system context.

        """
        # if run as root assume they want to install globally
        if os.geteuid() == 0:
            return cls.system_global

        # if python's environment prefix is writable --
        # likely a virtual environment --
        # default to installing alongside there
        if os.access(sys.prefix, os.W_OK):
            return cls.same_environ

        # finally -- safest -- default to the user's home path
        return cls.current_user

    @property
    def help_text(self):
        """verbose representation for help text"""
        help_text = self.description
        if self.target:
            help_text += f" ({self.target})"
        return help_text


def execute(argv=None, **parser_kwargs):
    """Execute the nPrint-installation CLI command."""
    path_option_default = PathOption.get_path_defaults()

    parser = argparse.ArgumentParser(
        description=(
            "install nPrint\n\n"

            "installation path:\n\n"

            "  Commands such as nPrint are frequently installed \"globally\" to a system – "
            "often under /usr/local/.\n\n"

            "  Specify any of the following options to install nPrint either:\n\n"

            + '\n'.join(f'    • {option}' for option in PathOption) + "\n\n" +

            "  For the current invocation and without options nPrint will be installed "
            f"by default {path_option_default.help_text}."
        ),
        formatter_class=argparse_formatter.FlexiFormatter,
        **parser_kwargs,
    )

    installation_path = parser.add_mutually_exclusive_group()
    installation_path.add_argument(
        '-p', '--prefix',
        metavar='PATH',
        help="specify an arbitrary installation path prefix "
             "(generally an alternative to /usr/local/)",
    )
    installation_path.add_argument(
        '-g', '--global',
        action='store_const',
        const=None,
        dest='prefix',
        help="install system-globally (generally under /usr/local/ – may require sudo)",
    )
    installation_path.add_argument(
        '-u', '--user',
        action='store_const',
        const=PathOption.current_user.target,
        dest='prefix',
        help=f"install for current user only (under {PathOption.current_user.target})",
    )
    installation_path.add_argument(
        '-s', '--same',
        action='store_const',
        const=PathOption.same_environ.target,
        dest='prefix',
        help=f"install into same environment as nprintML (under {PathOption.same_environ.target})",
    )
    installation_path.set_defaults(
        prefix=path_option_default.target,
    )

    args = parser.parse_args(argv)

    (build_commands, missing_commands) = get_build_commands(prefix=args.prefix)

    if missing_commands:
        parser.error('missing requirement(s): ' + ', '.join(missing_commands))

    missing_dependencies = get_missing_dependencies()

    if missing_dependencies:
        parser.error('missing dependenc(ies): ' + ', '.join(missing_dependencies) +
                     '\nconsult: https://github.com/nprint/nprint/wiki/2.-Installation')

    execute_bootstrap(build_commands)

    # check for pyenv tip
    try:
        pathlib.Path(args.prefix).relative_to(PYENV_PATH)
    except ValueError:
        # no apparent association with pyenv
        pass
    else:
        print('\nℹ nPrint appears to have been installed into a Pyenv environment – '
              'rehash may be required before it is available for use:\n\n'
              '\tpyenv rehash\n')


def get_build_commands(**kwargs):
    """Check for system availability of and return the system commands
    required to build nPrint.

    Command options may be populated via keyword argument. Unpopulated
    or null options are discarded.

    """
    context = {key: value for (key, value) in kwargs.items()
               if value is not None}

    def generate_command_args(args):
        for arg in args:
            try:
                yield arg.format_map(context)
            except KeyError:
                continue

    build_commands = [
        (shutil.which(cmd),) + tuple(generate_command_args(args))
        for (cmd, *args) in BUILD_COMMANDS
    ]

    missing_commands = {
        cmd for (found, cmd) in zip(
            (command[0] for command in build_commands),
            (command[0] for command in BUILD_COMMANDS),
        )
        if not found
    }

    return (
        build_commands,
        missing_commands,
    )


def get_missing_dependencies():
    """Look up dependencies & return those which could not be located."""
    whereis = shutil.which('whereis')

    if whereis is None:
        print('[warn] whereis command unavailable – '
              'cannot execute check for build dependencies',
              file=sys.stderr)

        return []

    dependency_results = (subprocess.check_output([whereis, lib]) for lib in BUILD_DEPENDENCIES)

    dependency_locations = (re.sub(rf'^{lib}: ?', '', result.decode())
                            for (lib, result) in zip(BUILD_DEPENDENCIES, dependency_results))

    try:
        return [
            lib for (lib, location) in zip(BUILD_DEPENDENCIES, dependency_locations)
            if not location.strip()
        ]
    except subprocess.CalledProcessError:
        print('[warn] whereis command failed – '
              'cannot execute check for build dependencies',
              file=sys.stderr)

        return []


def execute_bootstrap(commands):
    """Retrieve nPrint release & execute given installation commands."""
    with tempfile.TemporaryDirectory(prefix=f'{__name__}.') as tempdir:
        temp_path = pathlib.Path(tempdir)
        release_path = temp_path / NPRINT_RELEASE_NAME
        archive_path = temp_path / NPRINT_ARCHIVE_NAME

        urllib.request.urlretrieve(NPRINT_ARCHIVE_URL, archive_path)

        with tarfile.open(archive_path) as release_archive:
            release_archive.extractall(tempdir)

        for command in commands:
            print('→', *command, end='\n\n')

            try:
                subprocess.run(
                    command,
                    cwd=release_path,
                    check=True,
                )
            except subprocess.CalledProcessError as exc:
                print('\n✕ failed')
                sys.exit(exc.returncode)
            else:
                print()

    print('✓ success')


if __name__ == '__main__':
    execute()



import abc
import collections.abc
import importlib
import io
import pkgutil

import pandas as pd


class LabelAggregator(abc.ABC):
    """LabelAggregator provides general methods for working with nPrint
    data, such as attaching labels, compressing nPrint data, and
    generating new columns for flattened nPrint data.

    Inheritors of this class are expected to implement the abstract
    method `__call__()` to provide their method of feature aggregation.

    """
    def __init__(self, npt_csv, label_csv):
        self.npt_csv = npt_csv
        self.label_csv = label_csv

    @abc.abstractmethod
    def __call__(self, compress=False, sample_size=1):
        """Abstract method, expected to be implemented on a per-example
        label aggregation method.

        """

    def load_npt(self, npt_csv):
        """Load nPrint data.

        The index column is expected to *always* be the 0th column.

        """
        return pd.read_csv(npt_csv, index_col=0)

    def load_label(self, labels_csv):
        """Load labels, which are expected to be in item, label column
        format where item is the index.

        A header line "item,label" (case-insensitive) is optional.

        For example:

            item,label
            key1,label1
            key2,label2
            key3,label3

        """
        # enforce column names (item, label)
        label = pd.read_csv(labels_csv, index_col='item', names=('item', 'label'))

        # check for optional header line
        if not label.empty:
            row0 = label.iloc[0]
            if row0.name.lower() == 'item' and row0.label.lower() == 'label':
                return label.iloc[1:]

        return label

    def compress_npt(self, npt):
        """Compress columns out of an nPrint that provide no predictive
        signal.

        More specifically, this method drops *all* columns in a given nPrint
        dataframe where each bit is the same value.

        """
        nunique = npt.apply(pd.Series.nunique)
        cols_to_drop = nunique[nunique == 1].index
        return npt.drop(cols_to_drop, axis=1)

    def attach_label(self, npt, label):
        """Attach labels to a dataframe of nPrints, returning the labels
        that are missing, the new dataframe, and the number of samples
        that were lost.

        """
        samples0 = len(npt)
        missing_labels = npt.index.difference(label.index).tolist()

        try:
            npt = npt.join(label).dropna(subset=['label'])
        except KeyError as exc:
            raise LabelError("label input is malformed") from exc

        samples1 = len(npt)

        return (npt, missing_labels, samples0, samples1)

    def flatten_columns(self, columns, sample_size):
        """When we attach labels to more than one nPrint we need to
        create new columns that essentially are the original columns
        multiplied by the number of packets in each flattened sample.

        """
        return [
            f'pkt_{index}_{column}'
            for index in range(sample_size)
            for column in columns
        ]

    @staticmethod
    def filerepr(fd):
        return fd.name if isinstance(fd, io.TextIOWrapper) else str(fd)


class LabelError(Exception):
    pass


class PluginRegistry(collections.abc.Mapping):
    """Lazy class registry.

    The specified `package` is crawled -- lazily -- for well-named
    modules containing class-based plugins; and, as needed, subclasses
    of the specified `class_or_tuple` base (as with `issubclass()`)
    are imported and registered for retrieval.

    The registry operates as an immutable mapping.

    Values are imported plugin classes, stored under the names of the
    modules in which they are found.

    Plugin module discovery may be customized by overriding
    `__generate_names__`. Recognition and retrieval of plugin classes
    may be customized by overriding `__retrieve_member__`.

    """
    def __init__(self, class_or_tuple, package, ignore=()):
        # issubclass supports a class or a tuple of base classes, and so we do as well;
        # however, we also need to check for identity, and so we'll enforce type here:
        self.base = class_or_tuple if isinstance(class_or_tuple, tuple) else (class_or_tuple,)

        # at least ensure we have access to the base package
        self.package = importlib.import_module(package) if isinstance(package, str) else package

        self.ignore = ignore if isinstance(ignore, frozenset) else frozenset(ignore)

        self.__cache__ = None

    def __generate_names__(self):
        """Discover the plugin package's modules.

        Sub-packages and modules whose names are contained with `ignore`
        are ignored.

        This method should only be called internally and only *once*.
        It is factored out so as to be available for customization.

        """
        for module in pkgutil.iter_modules(self.package.__path__):
            if not module.ispkg and module.name not in self.ignore:
                yield module.name

    def __retrieve_member__(self, name):
        """Import the named module and return its contained plugin
        class.

        It is assumed that the module contains exactly one subclass of
        the `class_or_tuple` base.

        This method should only be called internally and only *once*.
        It is factored out so as to be available for customization.

        """
        module = importlib.import_module(f'{self.package.__name__}.{name}')
        (value,) = (member for member in vars(module).values()
                    if isinstance(member, type) and
                       issubclass(member, self.base) and  # noqa: E127
                       member not in self.base)           # noqa: E127
        return value

    def __populate_names__(self):
        """Construct and populate the keys of the registry cache if the
        cache does not exist.

        """
        if self.__cache__ is None:
            self.__cache__ = dict.fromkeys(self.__generate_names__())

    def __iter__(self):
        self.__populate_names__()
        yield from self.__cache__

    def __len__(self):
        self.__populate_names__()
        return len(self.__cache__)

    def __getitem__(self, name):
        self.__populate_names__()

        value = self.__cache__[name]

        if value is None:
            value = self.__cache__[name] = self.__retrieve_member__(name)

        return value



"""Pipeline Step to build models via AutoML: Learn"""
import argparse
import pathlib
import typing

from nprintml import pipeline

from . import AutoML


class LearnResult(typing.NamedTuple):
    """Pipeline Step results for Learn"""
    graphs_path: pathlib.Path
    models_path: pathlib.Path


class Learn(pipeline.Step):
    """Extend given `ArgumentParser` with the AutoML interface and
    build models from labeled `features`.

    Returns a `LearnResult`.

    """
    __requires__ = ('features',)

    def __init__(self, parser):
        group_parser = parser.add_argument_group(
            "automl",
        )

        group_parser.add_argument(
            '--test-size', '--test_size',
            default=AutoML.TEST_SIZE,
            metavar='FLOAT',
            type=NumericRangeType(float, (0, 1)),
            help="proportion of data split out for testing "
                 f"(float between zero and one defaulting to {AutoML.TEST_SIZE})",
        )
        group_parser.add_argument(
            '--metric',
            dest='eval_metric',
            default=AutoML.EVAL_METRIC,
            choices=AutoML.EVAL_METRICS_ALL,
            metavar='{...}',
            help=f"metric by which predictions will be evaluated (default: {AutoML.EVAL_METRIC})\n"
                 "select from:\n    " +
                 '\n    '.join(AutoML.EVAL_METRICS_ALL),
        )
        group_parser.add_argument(
            '-q', '--quality',
            default=AutoML.QUALITY,
            choices=range(len(AutoML.QUALITY_PRESETS)),
            metavar='INTEGER',
            type=int,
            help=f"model fit quality level (default: {AutoML.QUALITY})\n"
                 "select from:\n    " +
                 '\n    '.join(f'{index}: {label}'
                               for (index, label) in enumerate(AutoML.QUALITY_PRESETS)),
        )
        group_parser.add_argument(
            '--limit',
            default=AutoML.TIME_LIMITS,
            dest='time_limits',
            metavar='INTEGER',
            type=int,
            help="maximum time (seconds) over which to train each model "
                 f"(default: {AutoML.TIME_LIMITS})",
        )
        group_parser.add_argument(
            '--threads',
            default=AutoML.N_THREADS,
            dest='n_threads',
            metavar='INTEGER',
            type=NumericRangeType(int, (0, None)),
            help=f"number of CPU threads to dedicate to training (default: {AutoML.N_THREADS})",
        )

    def __call__(self, args, results):
        learn = AutoML(results.features, args.outdir / 'model')

        learn(
            test_size=args.test_size,
            eval_metric=args.eval_metric,
            quality=args.quality,
            time_limits=args.time_limits,
            n_threads=args.n_threads,
        )

        return LearnResult(
            graphs_path=learn.graphs_path,
            models_path=learn.models_path,
        )


class NumericRangeType:
    """Argument type validating that given value is of configured
    numeric type and bounds.

    Bounds may be specified as either `list` or `tuple`, the convention
    indicating whether bounds are "inclusive" or "non-inclusive",
    respectively.

    Either the lower or upper bound may be specified as `None`,
    indicating no bound.

    For example:

        parser.add_argument(
            '--test-size',
            default=0.3,
            metavar='FLOAT',
            type=NumericRangeType(float, (0, 1)),
        )

    The above argument, `--test-size`, will cast its input to `float`,
    (or print an error for non-float input). Moreover, a "not in range"
    error will be printed for inputs equal to zero or equal to or
    greater than one.

        parser.add_argument(
            '--threads',
            default=1,
            metavar='INTEGER',
            type=NumericRangeType(int, (0, None)),
        )

    In the above example, inputs are instead enforced as `int`; and,
    there is _only_ a lower bound -- inputs must only be greater than
    zero.

        parser.add_argument(
            '-q', '--quality',
            default=0,
            metavar='INTEGER',
            type=NumericRangeType(int, [0, 4]),
        )

    Arguments may also be given inclusive bounds, as above -- inputs to
    this example must be greater than or equal to zero, and less than or
    equal to four.

    """
    bounds_message = "upper and lower bounds must be list or tuple of two elements"

    def __init__(self, numeric_type, bounds):
        self.numeric_type = numeric_type

        try:
            (self.lower_bound, self.upper_bound) = bounds
        except ValueError as exc:
            raise ValueError(f"{self.bounds_message} not: {bounds!r}") from exc
        except TypeError as exc:
            raise TypeError(
                f"{self.bounds_message} not {bounds.__class__.__name__}: {bounds!r}"
            ) from exc

        if isinstance(bounds, tuple):
            self.inclusive = False
        elif isinstance(bounds, list):
            self.inclusive = True
        else:
            raise TypeError(f"{self.bounds_message} not {bounds.__class__.__name__}: {bounds!r}")

    def __call__(self, value):
        try:
            number = self.numeric_type(value)
        except ValueError:
            raise argparse.ArgumentTypeError(f"not {self.numeric_type.__name__}: {value!r}")

        if self.inclusive:
            if (
                (self.lower_bound is not None and number < self.lower_bound) or
                (self.upper_bound is not None and number > self.upper_bound)
            ):
                raise argparse.ArgumentTypeError(
                    f"not in range [{self.lower_bound}, {self.upper_bound}]: {number!r}"
                )
        else:
            if (
                (self.lower_bound is not None and number <= self.lower_bound) or
                (self.upper_bound is not None and number >= self.upper_bound)
            ):
                raise argparse.ArgumentTypeError(
                    f"not in range ({self.lower_bound}, {self.upper_bound}): {number!r}"
                )

        return number



"""Pipeline Step to aggregate nPrint data under labels appropriate for
machine learning: Label

"""
import argparse
import textwrap
import typing

import pandas as pd

from nprintml import pipeline

from .aggregator import registry as aggregators


class LabelResult(typing.NamedTuple):
    """Pipeline Step results for Label"""
    features: pd.DataFrame


class Label(pipeline.Step):
    """Extend given `ArgumentParser` with the label-aggregation
    interface and invoke the requested `LabelAggregator` to extract
    features from nPrint data for the subsequent pipeline step of
    machine learning.

    Returns a `LabelResult`.

    """
    __requires__ = ('nprint_path',)

    def __init__(self, parser):
        group_parser = parser.add_argument_group(
            "aggregation of data under supplied labels",
        )

        group_parser.add_argument(
            '-L', '--label-file', '--label_file',
            metavar='FILE',
            required=True,
            type=argparse.FileType('r'),
            help="label file (CSV)",
        )
        group_parser.add_argument(
            '-a', '--aggregator',
            choices=aggregators,
            required=True,
            help="label aggregation method",
        )
        group_parser.add_argument(
            '--help-aggregator',
            action=HelpAction,
            help_action=print_aggregators,
            help="describe aggregators and exit",
        )
        group_parser.add_argument(
            '--sample-size', '--sample_size',
            metavar='INTEGER',
            type=int,
            default=1,
        )
        group_parser.add_argument(
            '--compress',
            action='store_true',
            help="drop columns which do not appear to provide any predictive signal",
        )

    def __call__(self, args, results):
        aggregator_class = aggregators[args.aggregator]
        aggregator = aggregator_class(results.nprint_path, args.label_file)
        features = aggregator(compress=args.compress, sample_size=args.sample_size)
        return LabelResult(features)


def print_aggregators(parser, _namespace, _values, _option_string):
    """Document the aggregators by printing their doc strings."""
    print(f"{parser.prog}: labeling data aggregators")

    for (aggregator_name, aggregator_class) in aggregators.items():
        print('\n', aggregator_name, ':', sep='')

        # documentation
        doc = aggregator_class.__doc__ or aggregator_class.__init__.__doc__

        # fix indentation
        #
        # first line may not be indented at all
        if doc[0] in (' ', '\n'):
            doc = textwrap.dedent(doc)
        else:
            line_end = doc.find('\n') + 1
            if line_end > 1:
                doc = doc[:line_end] + textwrap.dedent(doc[line_end:])

        print(textwrap.indent(doc.strip(), '  '))


class HelpAction(argparse.Action):
    """Pluggable "help" action.

    Based on the built-in "help" action, enabling the definition of
    additional "--help-*" flags.

    Help output must be printed by the given callable `help_action`.

    When the flag is set, `help_action` is invoked and then the process
    is exited.

    """
    def __init__(self,
                 option_strings,
                 *,
                 help_action,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help=None):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

        self.help_action = help_action

    def __call__(self, parser, namespace, values, option_string=None):
        self.help_action(parser, namespace, values, option_string)
        parser.exit()



import pathlib

import pandas as pd

from . import LabelAggregator


class PcapLabelAggregator(LabelAggregator):
    """A LabelAggregator that takes a directory of nPrint output files
    and considers each file a single sample, padding to the maximum
    size, to run autoML on the generated samples.

    """
    def __init__(self, npt_csv, label_csv):
        if isinstance(npt_csv, str):
            npt_csv = pathlib.Path(npt_csv)

        if not isinstance(npt_csv, pathlib.Path):
            raise TypeError(
                f"{self.__class__.__name__} expects a path to a directory of nPrint "
                f"data file(s) not {npt_csv.__class__.__name__}: '{npt_csv}'"
            )

        if not any(npt_csv.iterdir()):
            raise FileNotFoundError(
                f"{self.__class__.__name__} requires at least one nPrint "
                f"data file but directory is empty: '{npt_csv}'"
            )

        super().__init__(npt_csv, label_csv)

    def __call__(self, compress=False, sample_size=1):
        """Enumerate given directory of nPrint data, load, pad to their
        maximum size, and attach labels.

        """
        npts = self.merge_npt()

        if compress:
            print('Compressing nPrint')
            npts = self.compress_npt(npts)
            print('  compressed nPrint shape:', npts.shape)

        print('Attaching labels to nPrints')
        labels = self.load_label(self.label_csv)

        (npt_df, missing_labels, ogns, nns) = self.attach_label(npts, labels)

        print('  labels attached: missing labels for:', missing_labels)
        print('    missing labels caused samples to be dropped:', (ogns - nns))

        return npt_df

    def load_label(self, labels_csv):
        labels = super().load_label(labels_csv)

        # Enable join on file path index with .npt-derived data --
        # strip any file suffix like .pcap (or .npt)
        #
        # (Unlike with former, which strictly loaded .pcap files, mapped these to .npt files, and
        # constructed its index from actual file paths, here we are loading file path values from
        # user-specified CSV data. As such, we won't assume too much about what suffix the user
        # supplied -- if any -- though .pcap is suggested.)
        #
        # Use regular expression rather than path method in case no extension specified, etc.,
        # (as intent is to strip it after all -- extension not meaningful):
        labels.index = labels.index.str.replace(r'\.(pcap|npt)$', '', case=False)

        return labels

    def merge_npt(self):
        """Merge nPrint data from multiple output files."""
        print('Loading nPrints')

        npts0 = []
        npt_paths = []
        largest_npt = None
        file_count = 0
        npt_files = (npt_file for npt_file in self.npt_csv.rglob('*') if npt_file.is_file())

        for (file_count, npt_file) in enumerate(npt_files, 1):
            npt = self.load_npt(npt_file)

            if largest_npt is None or npt.shape[0] > largest_npt.shape[0]:
                largest_npt = npt

            npts0.append(npt)
            npt_paths.append(str(npt_file.relative_to(self.npt_csv).with_suffix('')))

        print('Loaded', file_count, 'nprints')

        # We need every sample size to be the same, so we find the
        # largest nPrint in all of the files and pad each sample with
        # empty nPrints.

        row_len = npts0[0].shape[1] if npts0 else 0
        fill_row = [-1] * row_len

        npts1 = []

        for npt in npts0:
            for index in range(npt.shape[0], largest_npt.shape[0]):
                npt.loc[index] = fill_row
            npt_flat = npt.to_numpy().flatten()
            npts1.append(npt_flat)

        if largest_npt is None:
            cols_flat = ()
        else:
            cols_flat = self.flatten_columns(largest_npt.columns, largest_npt.shape[0])

        npts = pd.DataFrame(npts1, index=npt_paths, columns=cols_flat)

        print('nPrints padded to maximum size:',
              largest_npt if largest_npt is None else largest_npt.shape[0])
        print('nPrint shape:', npts.shape)

        return npts



"""nprintML command-line interface entry-point"""
import argparse
import itertools
import os
import pathlib
import re
import sys
import time

import argparse_formatter

import nprintml
from nprintml.pipeline import Pipeline

# ensure default steps of pipeline auto-load
import nprintml.net.step
import nprintml.label.step
import nprintml.learn.step


ANIMALS = ('aardvark', 'bison', 'canary', 'dalmation', 'emu', 'falcon', 'gnu',
           'hamster', 'impala', 'jellyfish', 'kiwi', 'lemur', 'manatee',
           'nutria', 'okapi', 'porcupine', 'quetzal', 'roadrunner', 'seal',
           'turtle', 'unicorn', 'vole', 'wombat', 'xerus', 'yak', 'zebra')


def execute(argv=None, **parser_kwargs):
    """Execute the nprintml CLI command."""
    try:
        parser = build_parser(**parser_kwargs)

        pipeline = Pipeline(parser)

        args = parser.parse_args(argv)

        check_output_directory(args)

        for (step, results) in pipeline(args):
            print(step, results, sep=' → ')

        print('done →', pipeline.results)
    except KeyboardInterrupt:
        print('interrupted ✕')
    except Exception as exc:
        if args.traceback:
            raise

        print(f'error:{exc_repr(exc)} ✕')
        sys.exit(1)


def build_parser(**parser_kwargs):
    """Construct base parser & arguments for nprintml.

    Link back to parser is set on any resulting parsed argument
    `Namespace` as `__parser__`.

    """
    parser = argparse.ArgumentParser(
        description='train models for network traffic analysis',
        formatter_class=argparse_formatter.FlexiFormatter,
        **parser_kwargs,
    )

    version = f'nprintML {nprintml.__version__} | nPrint {nprintml.__nprint_version__}'
    # support published command aliases
    prog = pathlib.Path(sys.argv[0])
    if prog.name != 'nprintml' and prog.suffix != '.py':
        version = f'%(prog)s {nprintml.__version__} | {version}'

    parser.add_argument(
        '--version',
        action='version',
        help="show program version and exit",
        version=version,
    )

    parser.add_argument(
        '-V', '--verbose',
        action='store_true',
        help="print human readable packets with nPrints",
    )

    parser.add_argument(
        '--tb', '--traceback',
        action='store_true',
        dest='traceback',
        help="print exception tracebacks",
    )

    output_default = get_default_directory()
    parser.add_argument(
        '-o', '--output',
        default=output_default,
        dest='outdir',
        metavar='DIR',
        type=pathlib.Path,
        help="output directory path to which to write artifacts and results "
             f"(default: {output_default})",
    )

    parser.set_defaults(
        __parser__=parser,
    )

    return parser


def exc_repr(exc):
    """Construct representation of given exception appropriate for
    printed output.

    """
    exc_repr = ''

    if exc.__class__.__module__ != 'builtins':
        exc_repr += f'{exc.__class__.__module__}.'

    exc_repr += exc.__class__.__name__

    if exc.args:
        exc_repr += ': ' + ', '.join(map(str, exc.args))

    return exc_repr


def pairwise(iterable):
    """s -> (s0, s1), (s1, s2), (s2, s3), ..., (sn, None)"""
    (a, b) = itertools.tee(iterable)
    next(b, None)
    return itertools.zip_longest(a, b)


def get_default_directory(base_name='nprintml', words=ANIMALS):
    """Construct user-friendly default output directory path."""
    path = pathlib.Path(base_name)

    if path.exists():
        word_options = '|'.join(words)
        run_pattern = re.compile(rf'run-({word_options})-', re.I)
        run_matches = (run_pattern.match(run_path.name) for run_path in path.glob('run-*'))
        words_used = sorted(
            (run_match.group(1).lower() for run_match in run_matches if run_match),
            reverse=True,
        )

        if words_used:
            last_word = words_used[0]
            for (word0, word1) in pairwise(words):
                if word0 == last_word:
                    next_word = word1 or words[0]
                    break
        else:
            next_word = words[0]
    else:
        next_word = words[0]

    return path / f'run-{next_word}-{int(time.time())}-{os.getpid()}'


def check_output_directory(args):
    """Ensure output directory exists and is empty."""
    if args.outdir.exists():
        if args.outdir.is_dir():
            if any(args.outdir.iterdir()):
                args.__parser__.error(f'output directory non-empty: {args.outdir}')
        else:
            args.__parser__.error(f'output path exists and is not a directory: {args.outdir}')
    else:
        args.outdir.mkdir(parents=True)



"""AutoML

Execute an AutoML procedure on a well-formated `DataFrame` containing a
`label` column. Uses AutoGluon.

Graphs may be constructed (with sane defaults) to visually illustrate
classifier performance.

"""
import itertools

import matplotlib.pyplot as plt
import seaborn as sns
from autogluon import TabularPrediction as task
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelBinarizer
from sklearn.metrics import (
    roc_curve,
    auc,
    precision_recall_curve,
    average_precision_score,
    confusion_matrix,
    ConfusionMatrixDisplay,
)


sns.set_style('ticks')
sns.set_context(rc={'lines.linewidth': 1.75})


class AutoML:
    """Execute AutoGluon on a `DataFrame` with a `label` column.

    Models and results are output to file.

    """
    linestyles = ('-', '--', '-.', ':')

    graphs_dirname = 'graphs'
    models_dirname = 'models'

    #: Parameter defaults and options
    EVAL_METRIC = 'accuracy'
    EVAL_METRICS_ALL = (
        'accuracy', 'balanced_accuracy', 'f1', 'f1_macro', 'f1_micro',
        'f1_weighted', 'roc_auc', 'average_precision', 'precision',
        'precision_macro', 'precision_micro', 'precision_weighted', 'recall',
        'recall_macro', 'recall_micro', 'recall_weighted', 'log_loss',
        'pac_score',
    )

    QUALITY = 0
    QUALITY_PRESETS = (
        'medium_quality_faster_train',
        'good_quality_faster_inference_only_refit',
        'high_quality_fast_inference_only_refit',
        'best_quality',
    )

    N_THREADS = 1

    TEST_SIZE = 0.3

    TIME_LIMITS = 5 * 60

    def __init__(self, data, outpath):
        self.data = data
        self.outpath = outpath

    @property
    def graphs_path(self):
        return self.outpath / self.graphs_dirname

    @property
    def models_path(self):
        return self.outpath / self.models_dirname

    def __call__(self, test_size=TEST_SIZE, eval_metric=EVAL_METRIC, quality=QUALITY,
                 time_limits=TIME_LIMITS, n_threads=N_THREADS):
        """Train, test, and evaluate models."""
        (train_data, test_data) = train_test_split(self.data, test_size=test_size)
        predictor = self.train(train_data, eval_metric, quality, time_limits,
                               n_threads)
        self.test(predictor, test_data)
        self.graph_all(predictor, test_data)

    def train(self, train_data, eval_metric=EVAL_METRIC, quality=QUALITY,
              time_limits=TIME_LIMITS, n_threads=N_THREADS):
        """Train prospective models."""
        # predictor gives us default access to the *best* predictor that
        # was trained on the task (otherwise we're just wrapping AutoGluon)
        return task.fit(train_data=train_data, label='label',
                        eval_metric=eval_metric,
                        output_directory=self.outpath,
                        time_limits=time_limits,
                        presets=self.QUALITY_PRESETS[quality],
                        nthreads_per_trial=n_threads)

    def test(self, predictor, test_data):
        """Evaluate models on the test set and write the results to file."""
        leaderboard = predictor.leaderboard(test_data, silent=True)
        leaderboard.set_index('model').sort_index().to_csv(self.outpath / 'leaderboard.csv')

    def graph_all(self, predictor, test_data):
        """Generate ROC, PR and confusion matrix graphs for the
        classification tasks.

        Uses sane defaults; (customization is left to the user with the
        models written to file).

        """
        y_true = test_data['label']
        test_no_label = test_data.drop(labels=['label'], axis=1)
        binarizer = LabelBinarizer().fit(y_true)
        binarized_labels = binarizer.transform(y_true)

        y_pred = predictor.predict(test_no_label)
        y_proba = predictor.predict_proba(test_no_label)
        self.make_pr(binarizer.classes_, binarized_labels, y_proba)
        self.make_roc(binarizer.classes_, binarized_labels, y_proba)
        self.make_cfmx(binarizer.classes_, y_true, y_pred)

    def make_cfmx(self, classes, y_true, y_pred):
        """Make confusion matrix without printing exact values.

        (Printing exact values of confusion can go poorly with large
        magnitudes of samples.)

        """
        cfmx = confusion_matrix(y_true, y_pred)
        cfmx_display = ConfusionMatrixDisplay(cfmx, display_labels=classes)
        cfmx_display.plot(include_values=False,
                          cmap=plt.cm.Blues,
                          xticks_rotation=45)

        self.print_graph('cfmx.pdf',
                         x_label='Predicted Label',
                         y_label='True Label')

    def make_pr(self, classes, y_true_bin, y_proba):
        """PR curve that splits on the binary case (no enumeration
        needed).

        """
        # Binary case does not enumerate
        if len(classes) == 2:
            plt_ax = self._make_binary_pr(y_true_bin.ravel(), y_proba.ravel())
        else:
            for i, class_label in enumerate(classes):
                plt_ax = self._make_binary_pr(y_true_bin[:, i], y_proba[:, i],
                                              class_label=class_label)

        # Split up the line styles to make them unique
        for (linestyle, line) in zip(itertools.cycle(self.linestyles),
                                     plt_ax.lines):
            line.set_linestyle(linestyle)

        self.print_graph('pr.pdf', x_lim=[0.0, 1.0],
                         y_lim=[0.0, 1.05], legend_loc='lower left',
                         x_label='Recall', y_label='Precision')

    def _make_binary_pr(self, y_true, y_proba, class_label=None):
        """Make a PR curve for the given data.

        Includes the average precision score for the given data.

        """
        (precision, recall, _) = precision_recall_curve(y_true, y_proba)
        cl_ap = average_precision_score(y_true, y_proba)

        if class_label:
            label = '{0} - AP: {1:.2f}'.format(class_label, cl_ap)
        else:
            label = 'AP: {0:.2f}'.format(cl_ap)

        return sns.lineplot(x=recall, y=precision, label=label)

    def make_roc(self, classes, y_true_bin, y_proba):
        """ROC curve that splits on the binary case (no enumeration
        needed).

        """
        # Binary case does not enumerate
        if len(classes) == 2:
            plt_ax = self._make_binary_roc(y_true_bin.ravel(), y_proba.ravel())
        else:
            for i, class_label in enumerate(classes):
                plt_ax = self._make_binary_roc(y_true_bin[:, i], y_proba[:, i],
                                               class_label=class_label)

        # Split up the line styles to make them unique
        for (linestyle, line) in zip(itertools.cycle(self.linestyles),
                                     plt_ax.lines):
            line.set_linestyle(linestyle)

        self.print_graph('roc.pdf', x_lim=[0.0, 1.0],
                         y_lim=[0.0, 1.05], x_label='False Positive Rate',
                         y_label='True Positive Rate',
                         legend_loc='lower right')

    def _make_binary_roc(self, y_true, y_proba, class_label=None):
        """Make ROC curve for given data.

        Includes roc_auc_score.

        """
        (fpr, tpr, _) = roc_curve(y_true, y_proba)
        cl_auc = auc(fpr, tpr)

        if class_label:
            label = '{0} - AUC: {1:.2f}'.format(class_label, cl_auc)
        else:
            label = 'AUC: {0:.2f}'.format(cl_auc)

        return sns.lineplot(x=fpr, y=tpr, label=label)

    def print_graph(self, ofn, x_lim=None, y_lim=None, legend_loc=None,
                    title=None, x_label=None, y_label=None):
        """Output graph."""
        plt.xlim(x_lim)
        plt.ylim(y_lim)

        if legend_loc:
            plt.legend(loc=legend_loc)

        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.title(title)
        plt.tight_layout()

        self.graphs_path.mkdir(parents=True, exist_ok=True)
        plt.savefig(self.graphs_path / ofn)

        plt.clf()



"""nprintml"""
import pathlib

from setuptools import find_packages, setup


README_PATH = pathlib.Path(__file__).parent / 'README.md'

INSTALL_REQUIRES = [
    'argparse-formatter == 1.2',

    'numpy ~= 1.19.4',
    'pandas ~= 1.1.4',

    # for cuda, e.g.: mxnet_cu101 < 2.0.0
    'mxnet < 2.0.0',

    # FIXME: Numpy pre-required in build environment to build Autogluon under py36-7 due to
    # mispackaging in its own dependency, ConfigSpace.
    #
    # Autogluon @ fa349db5e75a18cd3af7d9d3f1064eb34e92aca1 upgrades to a fixed ConfigSpace;
    # however, this revision has not yet been released to PyPI, and nprintML cannot be released to
    # PyPI with non-PyPI dependencies.
    #
    # As such, for the time being, numpy will be presumed in building nprintML, and nprintML
    # released as-is with the latest release of Autogluon.
    #
    # The Autogluon requirement should be upgraded as soon as possible, and the numpy dependency
    # removed from setup.cfg (tox).
    #
    # See nprintML @ 32ce3a2 for more information.
    #
    'autogluon ~= 0.0.14',
    'scikit-learn ~= 0.23.2',

    'matplotlib ~= 3.3.3',
    'seaborn ~= 0.11.0',
]

_DEV_REQUIRES = [
    'argcmdr==0.7.0',
    'bumpversion==0.6.0',
    'twine==3.2.0',
    'wheel==0.35.1',
]

_TESTS_REQUIRE = [
    'tox==3.20.1',
]

EXTRAS_REQUIRE = {
    'dev': _DEV_REQUIRES + _TESTS_REQUIRE,
    'test': _TESTS_REQUIRE,
}


setup(name='nprintml',
      version='0.0.3',
      description='Fully automated traffic analysis with nPrint',
      long_description=README_PATH.read_text(),
      long_description_content_type="text/markdown",
      url='https://github.com/nprint/nprintml',
      license='License :: OSI Approved :: Apache Software License',
      python_requires='>=3.6,<4',
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Telecommunications Industry',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Internet',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: System :: Networking :: Monitoring',
          'Topic :: Security',
          'Topic :: Terminals',
      ],
      packages=find_packages('src'),
      package_dir={'': 'src'},
      entry_points={
          'console_scripts': [
              'nml=nprintml.cli:execute',
              'nprintml=nprintml.cli:execute',
              'nprint-install=nprintml.net.install:execute',
            ],
      })



"""nprintML pipeline framework

The nprintML pipeline consists of concrete subclasses of `Step`,
serially executed by a `Pipeline`.

Steps may customize the CLI's `ArgumentParser` to ensure they receive
their required execution inputs. This is performed during construction
within the step's `__init__(parser)`.

Steps must define their executable actions, via definition of their
method `__call__(args, results)__`.

Steps may control their execution order, by specifying attribute names
of the `results` object, the values of which they require preceding
steps to determine. Required values are specified as a sequence of
attribute names at the step's class level attribute
`__requires__ = ('name0', 'name1', ...)`.

Steps may report their execution results by returning their own results
object from their `__call__`. (Generic objects and named tuples are
supported. Also consider the `dataclasses` decorator package.)

The pipeline is responsible for instantiating its steps (during its own
instantiation). Upon invocation, the pipeline returns a generator, which
invokes steps as their requirements are satisfied and merges their
results into the shared results object. Each iteration of this generator
yields a tuple consisting of the last executed step and its results (if
any).

The pipeline is itself a `set` of step instances as yet unexecuted.
Steps are removed from this set as they are invoked. The composite
results object is available as the pipeline instance attribute `results`
upon pipeline invocation.

A simple pipeline example follows:

    import argparse
    import pathlib
    import tempfile
    import typing
    import urllib.request
    from xml.etree import ElementTree as ET


    class GetDataResult(typing.NamedTuple):

        data_path: pathlib.Path


    class GetData(Step):

        def __init__(self, parser):
            parser.add_argument(
                '-u', '--url',
                default='https://www.lycos.com/',
                help="where's the data?",
            )

        def __call__(self, args, results):
            tempdir = tempfile.TemporaryDirectory(prefix=f'{__name__}.')
            data_path = pathlib.Path(tempdir) / 'data.html'
            urllib.request.urlretrieve(args.url, data_path)
            return GetDataResult(data_path)


    class UseDataResult(typing.NamedTuple):

        img_tags: list[ET.Element]


    class UseData(Step):

        __requires__ = ('data_path',)

        def __init__(self, parser):
            parser.add_argument(
                '-t', '--tag',
                default='img',
                help="tag type to extract",
            )

        def __call__(self, args, results):
            root = ET.parse(results.data_path)
            imgs = root.findall('.//img')
            return UseDataResult(imgs)


    if __name__ == '__main__':
        parser = argparse.ArgumentParser(description="scrape a web site")

        # pass base parser to Pipeline construction
        # (and thereby to Step constructors):
        pipeline = Pipeline(parser)

        args = parser.parse_args()

        # can exhaust pipeline iterator however --
        # e.g., gather step results into list or handle loop:
        for (step, result) in pipeline(args):
            print(step, result)

        # results collected on pipeline.results:
        print(pipeline.results)

In the above example, steps were automatically placed into a process-
global registry, from which the pipeline built itself by default.

In fact, any `set` may serve as a step registry, and a step may be added
to any number of registries. A step may be manually added to a registry
via `register`:

    registry = set()

    GetData.register(registry)

Alternatively, a step may be added to a custom registry upon
declaration:

    class SayHello(Step, registry=registry):

        def __call__(self, args, results):
            print('Hello')

And the pipeline may be contructed from any registry of steps:

    pipeline = Pipeline(parser, registry)

(The Step and Pipeline default registries may be overridden at the class/
metaclass level, via `__default_registry__`.)

"""
import abc
import collections
import types


#: Package default registry of concrete Steps
__step_registry__ = set()


class StepMeta(abc.ABCMeta):
    """Metaclass for `Step`."""
    #
    # Note that this metaclass doesn't do much of anything that *couldn't* be
    # handled on Step, (thanks to the modern hook __init_subclass__).
    #
    # However, abc entails a custom metaclass, either implicitly or explicitly;
    # and, this allows for cleaner encapsulation of class-specific
    # functionality, and a cleaner interface (as the metaclass interface is made
    # available to the class but *not* to the instance).
    #

    #: Registry to which concrete Steps will be added by default
    __default_registry__ = __step_registry__

    def register(cls, registry):
        """Add the Step to the given registry."""
        registry.add(cls)


class Step(metaclass=StepMeta):
    """Step of a Pipeline, specifying its command-line interface,
    executive functionality, and requirements from preceding Step(s).

    Steps are added to their `__default_registry__` by default upon
    declaration. A Step may be added to an alternate registry via
    declaration argument `registry`, e.g.:

        registry = set()

        class StepOne(Step, registry=registry):

            ...

    """
    #: Sequence of data keys on the results object -- gathered from the
    #: execution of preceding steps -- required for this step to execute
    __requires__ = ()

    @classmethod
    def __init_subclass__(cls, *, registry=None, **kwargs):
        """Register concrete Step upon declaration."""
        super().__init_subclass__(**kwargs)

        if getattr(cls, '__abstractmethods__', None):
            if registry is not None:
                raise TypeError(
                    f"Can't register abstract class {cls.__name__}. "
                    "To customize inherited default, override class attribute "
                    f"__default_registry__ (preferably on the metaclass {StepMeta.__name__}"
                )
        else:
            cls.register(cls.__default_registry__ if registry is None else registry)

    def __init__(self, parser):
        """Initialize Step & extend given `ArgumentParser` with Step's
        command-line interface.

        """

    @abc.abstractmethod
    def __call__(self, args, results):
        """Step's executive functionality.

        Args:
            args: parsed argument `Namespace`
            results: pipeline execution's composite results object
                (including results collected from all preceding steps)

        Return a structured data object to report the results of Step
        execution and to share these with subsequent Steps.

        """

    def __repr__(self):
        return 'step:' + self.__class__.__name__


class Pipeline(set):
    """Pipeline of executable Steps, instantiated from and extending an
    `ArgumentParser`, and executed upon iterative invocation.

    A `Pipeline` may instantiate any iterable of Steps. By default,
    those Steps registered under the pipeline's `__default_registry__`
    are used.

    The pipeline is itself a `set` of instantiated steps to be executed.
    Step execution order is determined at run-time, according to
    `Step.__requires__`, and those result keys which steps have
    returned. If step requirements cannot be met by a given pipeline,
    `StepRequirementError` is raised.

    Steps are removed from the pipeline collection as they are executed.

    Result objects returned by steps are merged with a composite
    pipeline result object, accessible at instance attribute `results`.

    A pipeline is executed by iterating over the result of its
    invocation -- either looping over this iterator or collecting it
    into a sequence.

    For example:

        parser = argparse.ArgumentParser(description="pipeline demo")
        pipeline = Pipeline(parser)
        args = parser.parse_args()

        # can exhaust pipeline iterator however --
        # e.g., gather step results into list or handle loop:
        for (step, result) in pipeline(args):
            print(step, result)

        # results collected on pipeline.results:
        print(pipeline.results)

    If neither iterative inspection nor collection of step results is
    desired, instance method `Pipeline.exhaust(args, results=None)`
    may be invoked rather than invoking the pipeline itself.

    For example, altering the above:

        pipeline.exhaust(args)

        print(pipeline.results)

    """
    #: Registry of Steps from which Pipeline will be constructed by default
    __default_registry__ = __step_registry__

    #: Default constructor for composite results object
    results_class = types.SimpleNamespace

    def __init__(self, parser, registry=None):
        """Construct a Pipeline of Steps extending the given
        `ArgumentParser`.

        """
        if registry is None:
            registry = self.__default_registry__

        super().__init__(step(parser) for step in registry)

        self.results = None

    def __call__(self, args, results=None):
        """Construct an iterator to execute the steps of the pipeline."""
        self.results = self.results_class() if results is None else results

        while self:
            results_available = set(self.results.__dict__.keys())
            steps_satisfied = [step for step in self
                               if results_available.issuperset(step.__requires__)]

            if not steps_satisfied:
                raise StepRequirementError(self, results_available)

            for step in steps_satisfied:
                step_results = step(args, self.results)
                if step_results is not None:
                    self.merge(step_results)
                self.remove(step)
                yield (step, step_results)

    def exhaust(self, args, results=None):
        """Execute the steps of the pipeline (non-iteratively)."""
        iterator = self(args, results)
        collections.deque(iterator, maxlen=0)

    def merge(self, results):
        """Copy the attributes of the given results object into the
        pipeline's composite results.

        """
        try:
            data = results.__dict__
        except AttributeError:
            data = results._asdict()

        for (key, value) in data.items():
            if hasattr(self.results, key):
                raise KeyError('result key already defined', key)

            setattr(self.results, key, value)


class PipelineError(Exception):
    pass


class StepRequirementError(PipelineError):

    message = "step requirements could not be satisfied"

    def __init__(self, steps, results_available):
        super().__init__(
            self.message,
            steps,
            results_available,
        )
        self.steps = steps
        self.results_available = results_available



"""The aggregator package defines concrete subclasses of
`LabelAggregator` for use in training, evaluating and representing the
results of AutoML training on nPrint data.

In general, these classes take a dataframe generated by the label
aggregation step of the nprintML pipeline. It can be used on its own as
long as it receives a dataframe where each row of the dataframe contains
a single sample and the labels for the samples are contained in a
`label` column in the dataframe.

Subclasses are organized into submodules, such as
`aggregator.index.IndexLabelAggregator`.

A lazy-loading `registry` of these classes is provided for their dynamic
inspection and retrieval.

"""
from .base import LabelAggregator, LabelError, PluginRegistry  # noqa: F401


registry = PluginRegistry(LabelAggregator,
                          __name__,
                          ignore=['base'])
